#ifndef _MATRIX_H_
#define _MATRIX_H_
#include "vector.h"

class Matrix
{
  public:
    double x[4][4];
    Matrix() { }
    Matrix(double);
    Matrix& operator +=(Matrix&);
    Matrix& operator -=(Matrix&);
    Matrix& operator *=(Matrix&);
    Matrix& operator *=(double);
    Matrix& operator /=(double);
    friend Matrix operator +(Matrix&, Matrix&);
    friend Matrix operator -(Matrix&, Matrix&);
    friend Matrix operator *(Matrix&, double);
    friend Matrix operator *(Matrix&, Matrix&);
    friend Vector operator *(Matrix&, Vector&);
};

Matrix Translate(Vector&);
Matrix Scale(Vector&);
Matrix RotateX(double);
Matrix RotateY(double);
Matrix RotateZ(double);
Matrix Rotate(Vector&, double);
Matrix MirrorX();
Matrix MirrorY();
Matrix MirrorZ();

Matrix::Matrix(double d)
{
  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
      x[i][j] = (i == j) ? d : 0.0;

  x[3][3] = 1;
}

Matrix& Matrix::operator +=(Matrix& A)
{
  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
      x[i][j] += A.x[i][j];

  return *this;
}

Matrix& Matrix::operator -=(Matrix& A)
{
  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
      x[i][j] -= A.x[i][j];

  return *this;
}

Matrix& Matrix::operator *=(Matrix& A)
{
  Matrix res = *this;

  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
    {
      double sum = 0;
      for (int k = 0; k < 4; k++)
        sum += res.x[i][k] * A.x[k][j];
      x[i][j] = sum;
    }

  return *this;
}

Matrix& Matrix::operator *=(double d)
{
  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
      x[i][j] *= d;

  return *this;
}

Matrix operator +(Matrix& A, Matrix& B)
{
  Matrix res;

  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
      res.x[i][j] = A.x[i][j] + B.x[i][j];

  return res;
}

Matrix operator -(Matrix& A, Matrix& B)
{
  Matrix res;

  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
      res.x[i][j] = A.x[i][j] - B.x[i][j];

  return res;
}

Matrix operator *(Matrix& A, Matrix& B)
{
  Matrix res;

  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
    {
      double sum = 0;
      for (int k = 0; k < 4; k++)
        sum += A.x[i][k] * B.x[k][j];
      res.x[i][j] = sum;
    }

  return res;
}

Matrix operator *(Matrix& A, double d)
{
  Matrix res;

  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
      res.x[i][j] = A.x[i][j] * d;

  return res;
}

Vector operator *(Matrix& M, Vector& v)
{
  Vector res;

  res.x = M.x[0][0] * v.x + M.x[0][1] * v.y + M.x[0][2] * v.z + M.x[0][3];
  res.y = M.x[1][0] * v.x + M.x[1][1] * v.y + M.x[1][2] * v.z + M.x[1][3];
  res.z = M.x[2][0] * v.x + M.x[2][1] * v.y + M.x[2][2] * v.z + M.x[2][3];

  double denom = M.x[3][0] * v.x + M.x[3][1] * v.y + M.x[3][2] * v.z + M.x[3][3];

  if (denom != 1.0)
    res /= denom;

  return res;
}

Matrix Translate(Vector& v)
{
  Matrix res(1);

  res.x[0][3] = v.x;
  res.x[1][3] = v.y;
  res.x[2][3] = v.z;

  return res;
}

Matrix Scale(Vector& v)
{
  Matrix res(1);

  res.x[0][0] = v.x;
  res.x[1][1] = v.y;
  res.x[2][2] = v.z;

  return res;
}

Matrix RotateX(double a)
{
  Matrix res(1);
  double cosine = cos(a);
  double sine = sin(a);

  res.x[1][1] = cosine;
  res.x[1][2] = -sine;
  res.x[2][1] = sine;
  res.x[2][2] = cosine;

  return res;
}

Matrix RotateY(double a)
{
  Matrix res(1);
  double cosine = cos(a);
  double sine = sin(a);

  res.x[0][0] = cosine;
  res.x[0][2] = -sine;
  res.x[2][0] = sine;
  res.x[2][2] = cosine;

  return res;
}

Matrix RotateZ(double a)
{
  Matrix res(1);
  double cosine = cos(a);
  double sine = sin(a);

  res.x[0][0] = cosine;
  res.x[0][1] = -sine;
  res.x[1][0] = sine;
  res.x[1][1] = cosine;

  return res;
}

Matrix Rotate(Vector& v, double a)
{
  Matrix res(1);
  double cosine = cos(a);
  double sine = sin(a);

  res.x[0][0] = v.x * v.x + (1 - v.x * v.x) * cosine;
  res.x[1][0] = v.x * v.y * (1 - cosine) + v.z * sine;
  res.x[2][0] = v.x * v.z * (1 - cosine) - v.y * sine;
  res.x[3][0] = 0;

  res.x[0][1] = v.x * v.y * (1 - cosine) - v.z * sine;
  res.x[1][1] = v.y * v.y + (1 - v.y * v.y) * cosine;
  res.x[2][1] = v.y * v.z * (1 - cosine) + v.x * sine;

  res.x[0][2] = v.x * v.z * (1 - cosine) + v.x * sine;
  res.x[1][2] = v.y * v.z * (1 - cosine) - v.x * sine;
  res.x[2][2] = v.z * v.z + (1 - v.z * v.z) * cosine;

  res.x[0][3] = 0;
  res.x[1][3] = 0;
  res.x[2][3] = 0;
  res.x[3][3] = 1;

  return res;
}

Matrix MirrorX()
{
  Matrix res(1);
  res.x[0][0] = -1;
  return res;
}

Matrix MirrorY()
{
  Matrix res(1);
  res.x[1][1] = -1;
  return res;
}

Matrix MirrorZ()
{
  Matrix res(1);
  res.x[2][2] = -1;
  return res;
}

#endif